/*
 * file: tcp_serverd.cc
 *
 * description:  this is the main module for the tcp_serverd project.
 *
 * revision list:
 *   Mon Jun 22 08:44:05 EDT 1998 Kyle R. Burton created the file,
 *       created skeleton fork, and sig handler code...
 *   Thu Jun 25 15:00:32 EDT 1998 KRB - changed from comm_server to
 *       tcp_serverd, started to add real functionality
 *   Fri Jun 26 11:22:28 EDT 1998 KRB - implemented sighup_handler,
 *       and got the fork/exec tied file descriptor thing working
 *       This is thanks to: Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 *       and his socket package: 
 *         ftp://atrey.karlin.mff.cuni.cz/pub/local/mj/linux/sock-*.tar.gz
 *       from which I lifted the dup() code.
 *   Mon Jul  6 07:56:42 EDT 1998 KRB - cleaned up code and directory
 *       in preperation for GPL'ing of sources...
 *   Mon Jul  6 09:25:32 EDT 1998 KRB - added sigint_handler so we can perform
 *       cleanup on ctrl+c
 *   Mon Jul  6 17:23:02 EDT 1998 KRB - fixed the log/config path stuff, and
 *       the problems with not using strdup/free on the global file names/paths,
 *       also added code so the globals are settable via either the config file,
 *       or command line options...
 *   Thu Jul  9 09:06:36 EDT 1998 KRB - I had made changes under HP-UX, and forgot
 *       to test them under Linux.  Of course they broke, so I fixed them.
 *   Thu Jul  9 10:55:47 EDT 1998 KRB - implemented the cmdline/conf file stuff
 *       for the buffering_type via setvbuf(3), and implemented the setvbuf(3) calls
 *       on the teid socket handels...I was enlightened of this technique by
 *       Edgar Bonet Orozco <bonet@labs.polycnrs-gre.fr> (thanks for the advice)
 *   Fri Jul 10 10:27:04 EDT 1998 KRB - finally got the logging stuff working
 *       correctly (dynamicly changeable between cout/and an ofstream).  Preparing
 *       for 'one dot oh' rsn.
 *   Thu Sep 10 07:51:06 EDT 1998 KRB - made changes so the server can restrict
 *       client connections to those initiated on the same machine.
 *
 */

// standard headers...
#include <iostream.h>
#include <fstream.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
// even though we're going to use iostream, we still
// need this for setvbuf(), sprintf(), remove(), and such...
#include <stdio.h>      
#include <errno.h>

// actually, this may be Linux specific
// anyone know about other platforms?
#if !defined( _HP_UX )
  #define _USE_BSD
  #include <sys/types.h>
  #include <sys/resource.h>
  #include <sys/wait.h>
#endif

// our 'standard' headers...
#include <stdinc.h>
#include <c_hash.h>
#include <c_timestamp.h>
#include <c_socket.h>
#include <c_auto_ptr.h>
#include <std_init.h>
#include <c_log_stream.h>

// project specific headers
#include <tcp_serverd.h>

// again, anyone aware how other platforms work?
#if defined _HP_UX
  #define SET_SIGHANDLER sigset
#else
  #define SET_SIGHANDLER signal
#endif


// handlers, and utility routines...
#if defined _SUN_OS
  void sighup_handler( int sig, siginfo_t *sip, void *uap );
  void sigint_handler( int sig, siginfo_t *sip, void *uap );
  void sigchld_handler( int sig, siginfo_t *sip, void *uap );
  void sigalrm_handler( int sig, siginfo_t *sip, void *uap );
#else
  void sighup_handler( int sig );               // reconfig on a hup
  void sigint_handler( int sig );               // clean up and die
  void sigchld_handler( int sig );              // reap child, close it's socket
  void sigalrm_handler( int sig );              // dump stats
#endif

int initialize_settings( void );              // read/parse the conf file
int parse_cmdline( int& argc, char*** argv ); // read/parse the cmdline args
void show_usage( void );                      // show help
int invoke_the_protocol_handler( c_tcp_socket& client );
int set_buffering_type( const char* type );

#if defined _SUN_OS
  struct sigaction sighup_sigaction;
  struct sigaction sigchld_sigaction;
  struct sigaction sigalrm_sigaction;
  struct sigaction sigint_sigaction;
#endif

// global settings and such...
c_hash<c_string,c_string> settings;          // this will hold cmdline/conf file settings
c_hash<c_string,c_string> protocol_handlers; // the map of protocols to commands
pid_t parent_pid = -1;                       // the main functions's pid
int     max_num_concurrent_children    = 25; 
int     current_num_children           = 0;
unsigned long clients_accepted         = 0;
unsigned long sighups                  = 0;
unsigned long sigchlds                 = 0;
unsigned long sigalrms                 = 0;
int     buffering_type                 = _IONBF;
char*   pid_file                       = (char*)NULL; // "tcp_server.pid";
char*   log_file                       = (char*)NULL; // "tcp_server.log";
char*   config_file                    = (char*)NULL; // "tcp_server.conf";
char*   scoreboard_file                = (char*)NULL; // "tcp_server.stats";
int     use_log_file                   = FALSE;
int     using_log_file                 = FALSE;
int     accepting_connections          = TRUE;
c_hash<pseudo_int,pseudo_int> child_pids(17);
int          scoreboard_frequency      = 600;
c_log_stream out;
ofstream     log_stream;
unsigned     short port                = 4137;
int          socket_bound              = FALSE;
int          default_buff_size         = 512;
char*        bin                       = (char*)"<no program name>";
int          main_argc                 = 0;
char**       main_argv                 = NULL;
c_tcp_socket server;
int restrict_clients_to_localhost      = FALSE;  // only allow connections from the
                                                 // same machine...


/*
 *  this is the business end of tcp_server
 *
 */
int main( int argc, char** argv )
{
  int len = default_buff_size;
  c_auto_ptr<char> thisaddr(default_buff_size);
  c_auto_ptr<char> peeraddr(default_buff_size);

  // set up defaults for some of the globals...
  log_file        = strdup("tcp_server.log");
  config_file     = strdup("tcp_server.conf");
  scoreboard_file = strdup("tcp_server.stats");
  pid_file        = strdup("tcp_server.pid");

  parent_pid = getpid();
  bin = *argv++; argc--;
  main_argc = argc; main_argv = argv;

  // perform init functions
  parse_cmdline( argc, &argv );

  // read the conf files...
  initialize_settings();

  ofstream fpid;
  fpid.open( pid_file );
  if( fpid.fail() ) {
    WARN(out) << "[main] Error opening file: " << pid_file << " for writing..." << endl;
  }
  else {
    fpid << parent_pid << endl;
    fpid.close();
  }

#if defined _SUN_OS
  memset(&sigint_sigaction,0,sizeof(struct sigaction));
  sigint_sigaction.sa_handler   = (void (*)(int))sigint_handler;
  if( -1 == sigaction(SIGINT,&sigint_sigaction,NULL) ) {
    ERROR(out) << "[main] Error setting signal handler for SIGINT." << endl;
    exit(1);
  }
  INFO(out) << "[main] sigint handler installed..." << endl;

  sigchld_sigaction.sa_handler   = (void(*)(int))sigchld_handler;
  if( -1 == sigaction(SIGCHLD,&sigchld_sigaction,NULL) ) {
    ERROR(out) << "[main] Error setting signal handler for SIGCHLD." << endl;
    exit(1);
  }
  INFO(out) << "[main] sigchld handler installed..." << endl;

  sighup_sigaction.sa_handler   = (void(*)(int))sighup_handler;
  if( -1 == sigaction(SIGHUP,&sighup_sigaction,NULL) ) {
    ERROR(out) << "[main] Error setting signal handler for SIGHUP." << endl;
    exit(1);
  }
  INFO(out) << "[main] sighup handler installed..." << endl;

  sigalrm_sigaction.sa_handler   = (void(*)(int))sigalrm_handler;
  if( -1 == sigaction(SIGALRM,&sigalrm_sigaction,NULL) ) {
    ERROR(out) << "[main] Error setting signal handler for SIGALRM." << endl;
    exit(1);
  }
  INFO(out) << "[main] sigalrm handler installed..." << endl;
#else
  // set up a handler for SIGINT
  if( SET_SIGHANDLER( SIGINT, sigint_handler ) == SIG_ERR ) {
    ERROR(out) << "[main] Error setting signal handler for SIGINT." << endl;
    ERROR(out) << strerror(errno) << endl;
    exit(1);
  }
  INFO(out) << "[main] sigint handler installed..." << endl;

  // set up a handler for SIGHUP
  if( SET_SIGHANDLER( SIGHUP, sighup_handler ) == SIG_ERR ) {
    ERROR(out) << "[main] Error setting signal handler for SIGHUP." << endl;
    exit(1);
  }
  INFO(out) << "[main] sighup handler installed..." << endl;

  // set up a handler for SIGCHLD
  if( SET_SIGHANDLER( SIGCHLD, sigchld_handler ) == SIG_ERR ) {
    ERROR(out) << "[main] Error setting signal handler for SIGCHLD." << endl;
    exit(1);
  }
  INFO(out) << "[main] sigchld handler installed..." << endl;

  // set up a handler for SIGALRM
  if( SET_SIGHANDLER( SIGALRM, sigalrm_handler ) == SIG_ERR ) {
    ERROR(out) << "[main] Error setting signal handler for SIGALRM." << endl;
    exit(1);
  }
  INFO(out) << "[main] sigalrm handler installed..." << endl;
#endif

  // dump the stats
  kill( parent_pid, SIGALRM );

  // set up the alarm so we do the stats on a regular basis...
  INFO(out) << "[main] setting alarm frequency to: " 
            << scoreboard_frequency << endl;
  alarm(scoreboard_frequency);

  // open and init the socket...
  pid_t child;
  c_tcp_socket client;

  // go into the accept loop...
  while( accepting_connections ) {
    if( !server.accept(client) ) { 
      WARN(out) << "[main] Error accepting connection.  This may have been caused " 
                << endl;
      WARN(out) << "[main] by a signal (like: SIGCHLD, SIGHUP, or SIGALRM)." << endl;
      continue;
    }

    ++clients_accepted;

    // verify the incomming address...
    peeraddr[0] = '\0';
    if( !client.getpeeraddr(peeraddr,len) ) {
      INFO(out) << "couldn't get client address." << endl;
      // hmm...if this fails, there's probably something wrong with
      // the socket itself...
      // client.send("Sorry, I couldn't determine your address.\n");
      // client.close();
      // continue;
    }

    thisaddr[0] = '\0';
    client.getsockaddr(thisaddr,len);
    INFO(out) << "client connected: " << peeraddr << "-->" << thisaddr << endl;

    // is this sufficient?  I think the call to thisaddr, and the
    // call to peeraddr will both end up as loopback if they're from the
    // same host -- but if a remote host connects, thisaddr will be the
    // actual ip address of the localhost, while peeraddr will be the ip
    // address of the remote host...but I'm not absolutely sure.

    if( restrict_clients_to_localhost ) {
      // drop it if it's not from the localhost...
      if( 0 != strcmp( thisaddr, peeraddr ) ) {
        client.send("Sorry, you're not on this host, and I "
                    "just can't deal with you right now.\n");
        client.close();
        continue;
      }
    }

    // added Thu Sep 10 11:49:16 EDT 1998 KRB - the check for max
    // num children...
    if( ++current_num_children >= max_num_concurrent_children ) {
      INFO(out) << "[main] max children [" << max_num_concurrent_children 
                << "] reached, aborting child process." << endl;
      client.close();
      continue;
    }

    // fork, and handle the child...
    if( -1 == (child = fork()) ) {
      ERROR(out) << "Error forking a child process, this connection is lost.";
      continue;
    }

    if( 0 == child ) {
      // this is the child...
      invoke_the_protocol_handler(client);
      ERROR(out) << "[main::child] You can't get here..." << endl;
      ERROR(out) << "[main::child] I can't handle this anymore." << endl;
      exit(1);
    }

    // the rest is in the parent...keep track of the child's pid,
    // and go wait for another connection...
    child_pids.add( child, client.get_socket_fd() );

    // this should only effect the parent's view of this socket...
    client.close();
  }

  INFO(out) << "[main] exiting..." << endl;

  // delete the pid file...
  remove( pid_file );

  return 0;
}

int initialize_settings( void )
{
  // read conf files -- use stdinit and such...
  if( !init_configure( config_file, settings ) ) {
    ERROR(out) << "Error, unable to read from conf file: " << config_file << endl;
    exit(1);
  }

  // get the path to the protocol handler file...
  c_string handler_file;
  c_map_pair<c_string,c_string> pair;
  pair.m_key = "protocol_handler_file";
  if( !settings.find( pair ) ) {
    ERROR(out) << "Error, unable to find the protocol handler file in the settings." 
               << endl;
    exit(1);
  }

  // get the protocol handler list...
  handler_file = pair.m_value;
  protocol_handlers.remove_all();
  if( !init_configure( handler_file, protocol_handlers ) ) {
    ERROR(out) << "Error reading the protocol handler file." << endl;
    exit(1);
  }

  c_string value;
  if( settings.find( c_string("max_num_concurrent_children"), value ) ) {
    max_num_concurrent_children = atoi(value);
    #if defined( _DEBUG )
      INFO(out) << "setting max_num_concurrent_children to " 
                << max_num_concurrent_children << endl;
    #endif
  }

  if( settings.find( c_string("use_log_file"), value ) ){
    if( 0 == strcasecmp( value, "true" ) ) {
      use_log_file = TRUE;
    }
    else if( 0 == strcasecmp( value, "false" ) ) {
      use_log_file = FALSE;
    }
    else {
      WARN(out) << "Error, unrecognized value while trying to set use_log_file: " 
                << value << "  Resetting use_log_file to FALSE." << endl;
      use_log_file = FALSE;
    }
  }

  if( settings.find( c_string("log_file"), value ) ) {
    free(log_file);
    log_file = strdup(value);
    #if defined( _DEBUG )
      INFO(out) << "setting log_file to " << log_file << endl;
    #endif
  }

  if( settings.find( c_string("config_file"), value ) ) {
    free(config_file);
    config_file = strdup(value);
    #if defined( _DEBUG )
      INFO(out) << "setting config_file to " << config_file << endl;
    #endif
  }

  if( settings.find( c_string("pid_file"), value ) ) {
    free(pid_file);
    pid_file = strdup(value);
    #if defined( _DEBUG )
      INFO(out) << "setting pid_file to " << pid_file << endl;
    #endif
  }

  if( settings.find( c_string("scoreboard_file"), value ) ) {
    free(scoreboard_file);
    scoreboard_file = strdup(value);
    #if defined( _DEBUG )
      INFO(out) << "setting scoreboard_file to " << scoreboard_file << endl;
    #endif
  }

  if( settings.find( c_string("scoreboard_frequency"), value ) ) {
    scoreboard_frequency = atoi(value);
    #if defined( _DEBUG )
      INFO(out) << "setting scoreboard_frequency to " << scoreboard_frequency << endl;
    #endif
  }

  unsigned short new_port = port;
  if( settings.find( c_string("port"), value ) ) {
    new_port = atoi(value);
  }

  #if defined( _DEBUG )
    INFO(out) << "setting port to " << port << endl;
  #endif
  if( new_port != port || !socket_bound ) {
    #if defined( _DEBUG )
      INFO(out) << "changing the port from: " << port << " to " << new_port << endl;
    #endif

    port = new_port;

    server.close();

    #if defined( _DEBUG )
      INFO(out) << "closed socket..." << endl;
    #endif

    if( c_tcp_socket::CREATED != server.create_socket() ) {
      ERROR(out) << "Error re-creating socket for port change " << new_port 
                 << " => " << port << endl;
      ERROR(out) << "this is a fatal error..." << endl;
      accepting_connections = FALSE;  // this causes the while(accept) to exit...
      return FALSE;
    }

    if( !server.bind(port) ) {
      ERROR(out) << "Error re-binding socket for port change " << new_port 
                 << " => " << port << endl;
      ERROR(out) << "this is a fatal error..." << endl;
      accepting_connections = FALSE;  // this causes the while(accept) to exit...
      return FALSE;
    }
  
    if( !server.listen() ) {
      ERROR(out) << "Error re-listening on socket for port change " << new_port 
                 << " => " << port << endl;
      ERROR(out) << "this is a fatal error..." << endl;
      accepting_connections = FALSE;  // this causes the while(accept) to exit...
    }

    socket_bound = TRUE;
  }

  if( settings.find( c_string("buffering_type"), value ) ) {
    set_buffering_type(value);
    #if defined( _DEBUG )
      INFO(out) << "setting set_buffering_type to " << value << endl;
    #endif
  }

  if( settings.find( c_string("restrict_clients_to_localhost"), value ) ) {
    if( 0 == strcasecmp( value, "true" ) ) {
      restrict_clients_to_localhost = TRUE;
    }
    else {
      restrict_clients_to_localhost = FALSE;
    }
  }

  if( use_log_file ) {
    out.use_file( log_file );
  }
  else {
    out.use_stdout();
  }

  return TRUE;
}


int parse_cmdline( int& argc, char*** _argv )
{
  char** argv = *_argv;

  // I should probably use getopts here...but I haven't learned it yet
  while( argc && argv[0][0] == '-' ) {
    switch( argv[0][1] ) {
      case 'B': // line buffering type...
        argc--; argv++;
        if( !argc ) {
          ERROR(out) << "Error, you must use either \"none\", \"full\", or "
                     << "\"line\" as a buffering type." << endl;
          exit(1);
        }
        set_buffering_type(*argv);
        break;
      case 'b': // buffer size
        argc--; argv++;
        if( !argc ) {
          ERROR(out) << "Error, you must specify a size with -b" << endl;
          exit(1);
        }
        default_buff_size = atoi(argv[0]);
        break;
      case 'c': // configuration file 
        argc--; argv++;
        if( !argc ) {
          ERROR(out) << "Erorr, you must specify a file with -c" << endl;
          exit(1);
        }
        free(config_file);
        config_file = strdup(argv[0]);
        break;
      case 'h': // show help
        show_usage();
        exit(1);
        break;
      case 'L': // use the log file instead of stdout...
        use_log_file = TRUE;
        break;
      case 'l': // log file
        argc--; argv++;
        if( !argc ) {
          ERROR(out) << "Error, you must speicfy a valid file with -l" << endl;
          exit(1);
        }
        free(log_file);
        log_file = strdup(*argv);
        break;
      case 'm': // max children
        argc--; argv++;
        if( !argc ) {
          INFO(out) << "Error, you must specify a maximum number with -m" << endl;
          exit(1);
        }
        max_num_concurrent_children = atoi(*argv);
        break;
      case 'P': // set the pid file
        argc--; argv++;
        if( !argc ) {
          ERROR(out) << "Error, you must specify a file with -P" << endl;
          exit(1);
        }
        pid_file = strdup(*argv);
        break;
      case 'p': // set the port
        argc--; argv++;
        if( !argc ) {
          INFO(out) << "Error, you must specify a port with -p" << endl;
          exit(1);
        }
        port = (unsigned short) atoi(*argv);
        break;
      case 's': // scoreboard file
        argc--; argv++;
        if( !argc ) {
          ERROR(out) << "Error, you must specify a file with -s" << endl;
          exit(1);
        }
        free(scoreboard_file );
        scoreboard_file = strdup(*argv);
        break;
      case 't': // scoreboard timeout
        argc--; argv++;
        if( !argc ) {
          INFO(out) << "Error, you must specify a timeout with -t" << endl;
          exit(1);
        }
        scoreboard_frequency = atoi(*argv);
        break;
      default:
        ERROR(out) << "Error, unrecognized cmdline swithc: " << argv[0] << endl;
        break;
    }
    argc--; argv++;
  }

  return TRUE;
}

void show_usage( void )
{
  cout << bin << " [opts]" << endl;
  cout << " Options are:" << endl;
  cout << "   -b size     set the default buffer size (def is 4096)." << endl;
  cout << "   -c file     use an alternate configuration file (def is tcpserver.conf)." 
       << endl;
  cout << "   -c path     use path as the loging directory (log file, and pid file)." 
       << endl;
  cout << "   -h          display this help screen." << endl;
  cout << "   -l file     use file as the log file." << endl;
  cout << "   -p port     use an alternate port to listen on (def is 4137)." << endl;
  cout << "   -P file     use file as the pid file." << endl;
  return;
}

#if defined _SUN_OS
  void sigchld_handler( int sig, siginfo_t *sip, void *uap )
#else
  void sigchld_handler( int sig )
#endif
{
  INFO(out) << "[sigchld_handler] one of life's greatest betryals is when you " 
            << "outlive your children..." << endl;

  ++sigchlds;
  --current_num_children;

  int status;
  pseudo_int socket_fd, child_id;
#if defined(_SUN_OS)
  // hold the sighandler...SUN OS uses unrelieable signals?
  sigset(SIGCHLD,SIG_HOLD);
  child_id = wait( &status );
#else
  child_id = wait3( &status, WNOHANG, NULL );
#endif

  if( -1 == child_id ) {
    ERROR(out) << "[sigchild_hadnler] Error retreiving child pid from wait3()." << endl;
    ERROR(out) << "[sigchild_hadnler] What the fsck! is going on here?" << endl;
    return;
  }

  if( !child_pids.find( child_id, socket_fd ) ) {
    ERROR(out) << "[sigchild_hadnler] Error retreiving child pid from child_pids?" 
               << endl;
    ERROR(out) << "[sigchild_hadnler] What the fsck! is going on here?" << endl;
    return;
  }

  child_pids.remove( child_id );
  close(socket_fd.val);
}

#if defined _SUN_OS
  void sighup_handler( int sig, siginfo_t *sip, void *uap )
#else
  void sighup_handler( int sig ) 
#endif
{
  INFO(out) << "[sighup_handler] the world changes around me, " << endl; 

  ++sighups;

  // re-read the config files
  initialize_settings();

  INFO(out) << "[sighup_handler] and still I remain the same." << endl; 
}

#if defined _SUN_OS
  void sigalrm_handler( int sig, siginfo_t *sip, void* uap )
#else
  void sigalrm_handler( int sig )
#endif
{
  // open the scoreboard file, and 
  // dump the stats to it...
  ofstream scoreboard;

  ++sigalrms;

  INFO(out) << "[sigalrm_handler] red alert." << endl;

  scoreboard.open( scoreboard_file, ios::out | ios::trunc );
  if( scoreboard.fail() ) {
    ERROR(out) << "[sigalrm_handler] couldn't open the scoreboard file." << endl;
    ERROR(out) << "[sigalrm_handler] no stats will be written for now..." << endl;
    return;
  }

   scoreboard << "Summary information: " << endl;
   scoreboard << "  max_num_concurrent_children:        " << max_num_concurrent_children 
              << endl;
   scoreboard << "  current_num_children:               " << current_num_children << endl;
   scoreboard << "  clients_accepted:                   " << clients_accepted << endl;
   scoreboard << "  log_file:                           " << log_file << endl;
   scoreboard << "  using_log_file:                     " << using_log_file << endl;
   scoreboard << "  pid_file:                           " << pid_file << endl;
   scoreboard << "  config_file:                        " << config_file << endl;
   scoreboard << "  scoreboard_frequency:               " << scoreboard_frequency << endl;
   scoreboard << "  default_buff_size:                  " << default_buff_size << endl;
   scoreboard << "  sighups:                            " << sighups << endl;
   scoreboard << "  sigchlds:                           " << sigchlds << endl;
   scoreboard << "  sigalrms:                           " << sigalrms << endl;
   scoreboard << endl;
 
   scoreboard.close();
 }

#if defined _SUN_OS
  void sigint_handler( int sig, siginfo_t *sip, void* uap )
#else
  void sigint_handler( int sig )
#endif
{
  // someone hit ctrl+c
  INFO(out) << "[sigint_handler] SIGINT caught, somebody's out to get us!" << endl;
  accepting_connections = FALSE;

  // this causes the accept() to return...and any further attempts to use it
  // to fail miserably...
  close( server.get_socket_fd() );
}

int invoke_the_protocol_handler( c_tcp_socket& client )
{
  c_string proto, handler;
  c_auto_ptr<char> buff(default_buff_size);
  c_map_pair<c_string,c_string> pair;

  // reset all the signals?

  // this close should only effect the child...
  server.close();

  int len = default_buff_size;
  if( !client.get_line( buff, len, '\n', '' ) ) {
    ERROR(out) << "Error, client didn't pass a legible protocol request." << endl;
    client.close();
    return FALSE;
  }
  proto = buff;

  INFO(out) << "client sent protocol request: \"" << proto << "\"" << endl;
  pair.m_key = proto;
  if( !protocol_handlers.find( pair ) ) {
    ERROR(out) << "Error, unable to find protocol handler for protocol: " 
               << proto << endl;
    client.send( "Error, you passed me an invlid protocol request: " );
    client.send( proto );
    return FALSE;
  }

  handler = pair.m_value;

  INFO(out) << "client requested protocol: " << proto << endl;
  INFO(out) << "calling handler:           " << handler << endl;

  int sk = client.get_socket_fd();
  INFO(out) << "socket fd is: " << sk << endl;

  // invoke some deep magic here...
  close(0); close(1); close(2);
  if(dup(sk) != 0 || dup(sk) != 1 || dup(sk) != 2 ) {
    ERROR(out) << "Error duping socket fd to stdin/stdout." << endl;
    return FALSE;
  }
  close(sk);

  setvbuf( stdin,  NULL, buffering_type, 0 );
  setvbuf( stdout, NULL, buffering_type, 0 );
  setvbuf( stderr, NULL, buffering_type, 0 );

  // is there a cleaner way do do this?
  execl( "/bin/sh", "/bin/sh", "-c", (const char*)handler, NULL );

  ERROR(out) << "Somebody pinch me so I know I'm not dreaming..." << endl;
  ERROR(out) << "You shouldn't be able to get here." << endl;
  exit(1);

  return TRUE;
}

int set_buffering_type( const char* type ) 
{
  c_string value(type);
  if( value == "line" ) {
    buffering_type = _IOLBF;
  }
  else if( value == "full" ) {
    buffering_type = _IOFBF;
  }
  else if( value == "none" ) {
    buffering_type = _IONBF;
  }
  else {
    WARN(out) << "unrecognized buffering type: " << type << " resetting to "
              << "default: \"none\"" << endl;
    buffering_type = _IONBF;
    return FALSE;
  }

  return TRUE;
}



